#include <stdio.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <sys/types.h>

#include "gfcc.h"

extern GtkWidget *notebook;
extern struct _fwchain fwchain[];
extern gint maxchains;
extern gint modified;

struct _fwentry fwentry;

static void edit_chain_enter(GtkWidget *, gint idx);
static void set_chain_value(GtkWidget *widget, gchar *data);
static void edit_chain(gchar *interface, gchar *chain, gint idx);
static void cwin_destroy(GtkWidget *, gpointer);

GtkWidget *chain_win = NULL;
ip_chainlabel newparent;
gint oldpage, oldrow;

ChainAction chainflag;

static void cwin_destroy(GtkWidget *widget, gpointer data)
{
	gtk_widget_destroy(chain_win);
	chain_win = NULL;
}

void edit_chain_window(GtkWidget *w, ChainAction flag)
{
	GtkWidget *vbox, *hbox;
	GtkWidget *table, *child;
	GtkWidget *label;
	GtkWidget *button, *option_menu, *submenu, *menu_item;
	gchar *text;
	gint i=0, j, height, idx;
	gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
	
	chainflag = flag;
	if(flag == CHILD) {
		find_chain(&idx, &oldrow, fwchain[page].label);
		height = 130;
	} else
		height = 90;

	if(chain_win)
		gtk_widget_destroy(chain_win);

	chain_win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_usize(chain_win, 235, height);
	gtk_window_set_position(GTK_WINDOW(chain_win), GTK_WIN_POS_CENTER);
	gtk_signal_connect(GTK_OBJECT(chain_win), "destroy",
			GTK_SIGNAL_FUNC(cwin_destroy), NULL);
	gtk_signal_connect(GTK_OBJECT(chain_win), "delete_event",
			GTK_SIGNAL_FUNC(cwin_destroy), NULL);
	if(flag == PARENT)
		gtk_window_set_title(GTK_WINDOW(chain_win), "New Chain");
	else
		gtk_window_set_title(GTK_WINDOW(chain_win), "Edit Chain");
	gtk_container_set_border_width(GTK_CONTAINER(chain_win), 3);
	
	vbox = gtk_vbox_new(FALSE, 1);
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(chain_win), vbox);
	
	table = gtk_table_new(3, 2, FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(table), 5);
	gtk_container_add(GTK_CONTAINER(vbox), table);
	gtk_table_set_col_spacing(GTK_TABLE(table), 0, 8);
	gtk_widget_show(table);
	
	if(flag == CHILD) {
		child = gtk_alignment_new(1,1,0,1);
		gtk_widget_show(child);
		gtk_table_attach_defaults(GTK_TABLE(table), child, 0,1,i,i+1);
		label = gtk_label_new("Parent:");
		gtk_widget_show(label);
		gtk_container_add(GTK_CONTAINER(child), label);
		
		option_menu = gtk_option_menu_new();
		gtk_widget_show(option_menu);
		submenu = gtk_menu_new();
		strcpy(newparent, fwchain[idx].label);
		menu_item = append_submenu_item(submenu,
			GTK_SIGNAL_FUNC(set_chain_value), fwchain[idx].label);
		if(strcmp(fwchain[idx].label, IP_FW_LABEL_INPUT))
			menu_item = append_submenu_item(submenu,
				GTK_SIGNAL_FUNC(set_chain_value),
				IP_FW_LABEL_INPUT);
		if(strcmp(fwchain[idx].label, IP_FW_LABEL_FORWARD))
			menu_item = append_submenu_item(submenu,
				GTK_SIGNAL_FUNC(set_chain_value),
				IP_FW_LABEL_FORWARD);
		if(strcmp(fwchain[idx].label, IP_FW_LABEL_OUTPUT))
			menu_item = append_submenu_item(submenu,
				GTK_SIGNAL_FUNC(set_chain_value),
				IP_FW_LABEL_OUTPUT);
		gtk_option_menu_set_menu(
				GTK_OPTION_MENU(option_menu), submenu);

		gtk_table_attach_defaults(
				GTK_TABLE(table), option_menu, 1,2,i,i+1);

		i++;
	}

	child = gtk_alignment_new(1,1,0,1);
	gtk_widget_show(child);
	gtk_table_attach_defaults(GTK_TABLE(table), child, 0,1,i,i+1);
	label = gtk_label_new("Name:");
	gtk_widget_show(label);
	gtk_container_add(GTK_CONTAINER(child), label);
	fwentry.label = gtk_entry_new_with_max_length(sizeof(ip_chainlabel)-1);
	gtk_table_attach_defaults(GTK_TABLE(table), fwentry.label, 1,2,i,i+1);
	if(flag == CHILD) {
		gtk_entry_set_text(
			GTK_ENTRY(fwentry.label), fwchain[page].label);
	}
	gtk_widget_show(fwentry.label);
	i++;
	
	child = gtk_alignment_new(1,1,0,1);
	gtk_widget_show(child);
	gtk_table_attach_defaults(GTK_TABLE(table), child, 0,1,i,i+1);
	label = gtk_label_new("Interface:");
	gtk_widget_show(label);
	gtk_container_add(GTK_CONTAINER(child), label);
	fwentry.via = gtk_entry_new_with_max_length(IFNAMSIZ-1);
	gtk_table_attach_defaults(GTK_TABLE(table), fwentry.via, 1,2,i,i+1);
	if(flag == CHILD) {
		gtk_clist_get_text(
			GTK_CLIST(fwchain[idx].chainlist), oldrow, 5, &text);
		oldpage = idx;
		gtk_entry_set_text(GTK_ENTRY(fwentry.via), text);
	}
	gtk_widget_show(fwentry.via);

	hbox = gtk_hbox_new(FALSE, 1);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 7);
	gtk_widget_show(hbox);

	button = gtk_button_new_with_label("OK");
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 1);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(edit_chain_enter), (gpointer)idx);
	gtk_widget_show(button);
	button = gtk_button_new_with_label("Cancel");
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 1);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(cwin_destroy), NULL);
	gtk_widget_show(button);

	gtk_widget_show(chain_win);
}

void set_chain_value(GtkWidget *w, gchar *data)
{
	strcpy(newparent, data);
}

void delete_chain(GtkWidget *w, gchar *data)
{
	gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
	gint i, j;
	
	find_chain(&i, &j, fwchain[page].label);
	
	if(GTK_TOGGLE_BUTTON(w)->active) {
		fwchain[page].deleted = 1;
		gtk_clist_set_text(GTK_CLIST(fwchain[i].chainlist),j,18,"1");
	} else {
		fwchain[page].deleted = 0;
		gtk_clist_set_text(GTK_CLIST(fwchain[i].chainlist),j,18,"0");
	}
	modified = 1;
}

void edit_chain_enter(GtkWidget *w, gint idx)
{
	gchar *interface, *chain;
	
	interface = gtk_entry_get_text(GTK_ENTRY(fwentry.via));
	if(!*interface)
		return;
	chain = gtk_entry_get_text(GTK_ENTRY(fwentry.label));
	if(!*chain)
		return;
	
	edit_chain(interface, chain, idx);

	gtk_widget_destroy(chain_win);
	chain_win = NULL;
}

void edit_chain(gchar *interface, gchar *chain, gint idx)
{
	gint i, j, page;
	static gchar *content[NUMCOLS], *label, *via;
	gchar savinterface[IFNAMSIZ];
	ip_chainlabel savchain;
	gchar *text;

	if(chainflag == PARENT)
		page = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
	else {
		for(page=0; page<maxchains; page++) {
			if(fwchain[page].policy[0] == '-')
				continue;
			if(!strcmp(fwchain[page].label, newparent))
				break;
		}
		gtk_clist_get_text(GTK_CLIST(fwchain[oldpage].chainlist),
				oldrow, 5, &text);
		strcpy(savinterface, text);
		gtk_clist_get_text(GTK_CLIST(fwchain[oldpage].chainlist),
				oldrow, 6, &text);
		strcpy(savchain, text);
		gtk_clist_remove(GTK_CLIST(fwchain[oldpage].chainlist),oldrow);
	}
	for(i=0; i < GTK_CLIST(fwchain[page].chainlist)->rows; i++) {
		gtk_clist_get_text(
			GTK_CLIST(fwchain[page].chainlist),i,5, &via);
		if(!strcmp(via, interface)) {
			dialog_window(
				"Parent Chain and Via name violation", NULL);
			page = oldpage;
			goto RECOVER;
		}
	}
	if(find_chain(&i, &j, chain)) {
		dialog_window("Chain name violation", NULL);
		page = oldpage;
		goto RECOVER;
	}
	
	modified = 1;
	strcpy(savinterface, interface);
	strcpy(savchain, chain);
		
RECOVER:
	content[0] = content_new("Any");
	content[2] = content_new("Any");
	content[5] = content_new(savinterface);
	content[6] = content_new(savchain);
	content[7] = content_new("FF");
	content[8] = content_new("00");
	content[9] = content_new("0");
	content[10] = content_new("65535");
	content[11] = content_new("0");
	content[12] = content_new("65535");
	content[13] = content_new("0");
	content[18] = content_new("0");

	gtk_clist_append(GTK_CLIST(fwchain[page].chainlist), content);

	fwchain[maxchains].refcnt = 1;
	strcpy(fwchain[maxchains].label, chain);
	strcpy(fwchain[maxchains].policy, "-");

	if(chainflag == CHILD) {
		page = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
		strcpy(fwchain[page].label, savchain);
	} else {
		page = maxchains++;
		draw_chain(notebook, page);
	}
	gtk_label_set_text(GTK_LABEL(fwchain[page].tablabel), chain);
	show_interface_desc(page, idx);
	
	g_free(content[0]);
	g_free(content[2]);
	for(i=5; i<14; i++)
		g_free(content[i]);
	g_free(content[18]);
}
