#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <sys/types.h>

#include "gfcc.h"

extern GtkWidget *window;
extern GtkWidget *notebook;

static gchar *titles[NUMCOLS] = {
	" Source",	" port",	" Destination",	" port",
	" Proto",	"Interface",	"Rule",		" A ",
	" X ",		"portsmin",	"portsmax",	"portdmin",
	"portdmax",	"fw_flg",	"fw_invflg",	"redirect",
	"fw_mark",	"outputsize",	"deleted",
};

struct _menu service_tcp[100];
struct _menu service_udp[100];
struct _menu proto_name[20];

struct _menu rule_name[10] = {
	{ 1, IP_FW_LABEL_ACCEPT },
	{ 1, IP_FW_LABEL_BLOCK },
	{ 1, IP_FW_LABEL_MASQUERADE },
	{ 1, IP_FW_LABEL_REJECT },
	{ 1, IP_FW_LABEL_RETURN },
	{ 1, IP_FW_LABEL_REDIRECT },
	{ 1, IP_FW_LABEL_NONE },
	{ 0, NULL },
};

struct _menu tos_name[10] = {
	{ 0xff, "0xFF - Not Set" },
	{ 0x10, "0x10 - Minimise Delay" },
	{ 0x08, "0x08 - Minimise Throughput" },
	{ 0x04, "0x04 - Maximise Reliability" },
	{ 0x01, "0x01 - Minimise Cost" },
	{ 0x00, "Normal Service" },
	{    0, NULL },
};

struct _fwchain fwchain[MAXCHAINS];
GtkWidget *cliplist = NULL;
gint maxchains=0;
gint modified=0;

static gint openpurpose;
static gchar *content[NUMCOLS];
static gchar fbuf[256];
static GtkWidget *popup_menu = NULL;

static gint rulelist_event_handler(GtkWidget *, GdkEvent *, gpointer);
static GtkWidget *create_rule_popup_menu(void);
static void set_fwpolicy_data(GtkWidget *widget, gpointer data);
static void get_fwnames(gchar *filename);
static void draw_default_chain();
static void draw_policy(gint idx);
static void get_fwrules(gchar *filename);
static void put_to_rulelist(struct _fwrule);
static void put_unvisible_data(struct _fwrule tfw);
static void clearChain();
static void file_select(void);
static void file_ok_sel(GtkWidget *, GtkWidget *);
static void file_can_sel(GtkWidget *, GtkWidget *);
static void file_destroy(GtkWidget *, GtkWidget *);
static void write_to_file(FILE *, struct ip_fwuser, ip_chainlabel);
static void force_load_file(GtkWidget *, gpointer);
static void force_load_system(GtkWidget *, gpointer);

static gchar *current_filename = NULL;
static gint set_and_quit;

void get_fw_data(gchar *filename, gint flag)
{
	gint i;
	gchar buf[128];

	set_and_quit = flag;

	if(set_and_quit) {
		if(!clear_system()) {
			printf("%s\n", ipfwc_strerror(errno));
			return;
		}
	}

	current_filename = filename;
	get_fwnames(filename);
	get_fwrules(filename);
	
	if(set_and_quit) {
		return;
	}

	if(filename)
		sprintf(buf, "%s: %s", PACKAGE_VERSION,filename);
	else
		sprintf(buf, "%s: %s", PACKAGE_VERSION,"Current System rules");

	gtk_window_set_title(GTK_WINDOW (window), buf);

	for(i=0; i<maxchains; i++)
		show_interface_desc(i, -1);
	
	if(cliplist) {
		gtk_clist_clear(GTK_CLIST(cliplist));
	} else
		cliplist = gtk_clist_new(NUMCOLS);

	modified = 0;
}

static void get_fwnames(gchar *filename)
{
	FILE *fp;
	gint i, ret=1;
	
	if(filename) {
		if(!(fp = fopen(filename, "r"))) {
			perror(filename);
			exit(1);
		}
	} else if(!(fp = fopen("/proc/net/"IP_FW_PROC_CHAIN_NAMES, "r"))) {
		fprintf(stderr, "Your kernel doesn't support ipchains\n");
		exit(2);
	}

	if(filename) {
		fgets(fbuf, sizeof(fbuf), fp);
		if(memcmp(fbuf, "#!ip-firewall", 13)) {
			fclose(fp);
			dialog_window("May be It's not a data file", NULL);
			return;
		}
	}
	clearChain();
	while(fgets(fbuf, sizeof(fbuf), fp)) {
		if(fbuf[0] == '#' || fbuf[0] == '\n')
			continue;
		if(fbuf[0] == '*')
			break;
		sscanf(fbuf, "%s %s %u %u %u %u %u",
			 fwchain[maxchains].label,   fwchain[maxchains].policy,
			&fwchain[maxchains].refcnt, &fwchain[maxchains].pkthi,
			&fwchain[maxchains].pktlo,  &fwchain[maxchains].bytehi,
			&fwchain[maxchains].bytelo);
		
		if(!set_and_quit)
			draw_chain(notebook, maxchains);
		else {
			if(strcmp(fwchain[maxchains].label,
						IP_FW_LABEL_INPUT)&&
			   strcmp(fwchain[maxchains].label,
						IP_FW_LABEL_FORWARD) &&
			   strcmp(fwchain[maxchains].label,
						IP_FW_LABEL_OUTPUT)) {
				ret = ipfwc_create_chain(
						fwchain[maxchains].label);
			} else {
				ret = ipfwc_set_policy(
						fwchain[maxchains].label,
						fwchain[maxchains].policy);
			}
		}
		if(!ret) {
			fprintf(stderr, "%s\n", ipfwc_strerror(errno));
			fclose(fp);
			return;
		}
		maxchains++;
	}
	fclose(fp);
	
	if(!maxchains && !set_and_quit)
		draw_default_chain();
}

static void draw_default_chain()
{
	strcpy(fwchain[0].label, IP_FW_LABEL_INPUT);
	strcpy(fwchain[0].policy, IP_FW_LABEL_ACCEPT);
	draw_chain(notebook, 0);

	strcpy(fwchain[1].label, IP_FW_LABEL_FORWARD);
	strcpy(fwchain[1].policy, IP_FW_LABEL_ACCEPT);
	draw_chain(notebook, 1);

	strcpy(fwchain[2].label, IP_FW_LABEL_OUTPUT);
	strcpy(fwchain[2].policy, IP_FW_LABEL_ACCEPT);
	draw_chain(notebook, 2);
		
	maxchains = 3;
}

void draw_chain(GtkWidget *notebook, gint idx)
{
	GtkWidget *vbox, *button;
	GtkWidget *scrolled_window;
	gint i;

	if(!popup_menu)
		popup_menu = create_rule_popup_menu();
	vbox = gtk_vbox_new(FALSE, 1);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
	gtk_widget_show(vbox);
		
	fwchain[idx].hbox = gtk_hbox_new(FALSE, 1);
	gtk_widget_show(fwchain[idx].hbox);
	gtk_box_pack_start(GTK_BOX(vbox),
			fwchain[idx].hbox, FALSE, FALSE, 2);

	button = gtk_button_new_with_label(" Zero counter ");
	gtk_box_pack_end(GTK_BOX(fwchain[idx].hbox), button, FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(set_zero_entries), NULL);
	gtk_widget_show(button);

	if(fwchain[idx].policy[0] != '-')
		draw_policy(idx);

	scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
			GTK_SCROLLED_WINDOW(scrolled_window),
			GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(vbox), scrolled_window);
	gtk_widget_show(scrolled_window);
	
	fwchain[idx].rulelist = gtk_clist_new_with_titles(NUMCOLS, titles);
	gtk_clist_set_selection_mode(
		GTK_CLIST(fwchain[idx].rulelist), GTK_SELECTION_EXTENDED);
	gtk_signal_connect(GTK_OBJECT(fwchain[idx].rulelist), "event",
				GTK_SIGNAL_FUNC(rulelist_event_handler), NULL);
	for(i=9; i<NUMCOLS; i++) {
		gtk_clist_set_column_visibility(
				GTK_CLIST(fwchain[idx].rulelist),i,FALSE);
	}
	gtk_clist_set_column_width(GTK_CLIST(fwchain[idx].rulelist), 0, 110);
	gtk_clist_set_column_width(GTK_CLIST(fwchain[idx].rulelist), 1, 85);
	gtk_clist_set_column_width(GTK_CLIST(fwchain[idx].rulelist), 2, 110);
	gtk_clist_set_column_width(GTK_CLIST(fwchain[idx].rulelist), 3, 85);
	gtk_clist_set_column_width(GTK_CLIST(fwchain[idx].rulelist), 4, 60);
	gtk_clist_set_column_width(GTK_CLIST(fwchain[idx].rulelist), 6, 65);
	gtk_scrolled_window_add_with_viewport(
		GTK_SCROLLED_WINDOW(scrolled_window), fwchain[idx].rulelist);
	gtk_widget_show(fwchain[idx].rulelist);
	
	fwchain[idx].chainlist = gtk_clist_new(NUMCOLS);

	fwchain[idx].tablabel = gtk_label_new(fwchain[idx].label);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
				vbox, fwchain[idx].tablabel);
}

void set_zero_entries(GtkWidget *widget, gpointer data)
{
	gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
	gchar buf[64];
	
	if(!ipfwc_zero_entries(fwchain[page].label))
		sprintf(buf, "%s.\n\nDid you have apply this data?",
				(gchar *)ipfwc_strerror(errno));
	else
		strcpy(buf, "Successfully zeroed the counters\nin this chain");
	
	dialog_window(buf, NULL);
}

static void draw_policy(gint idx)
{
	GtkWidget *button;

	button = gtk_radio_button_new_with_label(NULL, IP_FW_LABEL_ACCEPT);
	gtk_box_pack_start(GTK_BOX(fwchain[idx].hbox), button, FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(set_fwpolicy_data),
			IP_FW_LABEL_ACCEPT);
	if(!strcmp(fwchain[idx].policy, IP_FW_LABEL_ACCEPT))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
	gtk_widget_show(button);

	button = gtk_radio_button_new_with_label(
			gtk_radio_button_group(GTK_RADIO_BUTTON(button)),
			IP_FW_LABEL_BLOCK);
	gtk_box_pack_start(GTK_BOX(fwchain[idx].hbox), button, FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(set_fwpolicy_data), IP_FW_LABEL_BLOCK);
	if(!strcmp(fwchain[idx].policy, IP_FW_LABEL_BLOCK))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
	gtk_widget_show(button);

	button = gtk_radio_button_new_with_label(
			gtk_radio_button_group(GTK_RADIO_BUTTON(button)),
			IP_FW_LABEL_REJECT);
	gtk_box_pack_start(GTK_BOX(fwchain[idx].hbox), button, FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(set_fwpolicy_data),
			IP_FW_LABEL_REJECT);
	if(!strcmp(fwchain[idx].policy, IP_FW_LABEL_REJECT))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
	gtk_widget_show(button);
	
	button = gtk_button_new_with_label(" New Chain ");
	gtk_box_pack_end(GTK_BOX(fwchain[idx].hbox), button, FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(edit_chain_window), (ChainAction *)PARENT);
	gtk_widget_show(button);
}

static void get_fwrules(gchar *filename)
{
	FILE *fp = NULL;
	struct _fwrule tfw;
	struct ip_fwuser sfw;
	__u32 pkthi, pktlo, bytehi, bytelo;
	gint nread;

	if(filename)
		fp = fopen(filename, "r");
	else
		fp = fopen("/proc/net/"IP_FW_PROC_CHAINS, "r");
	if(!fp)
		return;

	if(filename) {
		while(fgets(fbuf, sizeof(fbuf), fp)) {
			if(fbuf[0] == '*')
				break;
		}
	}
	while(fgets(fbuf, sizeof(fbuf), fp)) {
		if(fbuf[0] == '#' || fbuf[0] == '\n')
			continue;
		if(fbuf[0] == '*')
			break;
		sscanf(fbuf, "%s "
			     "%X/%X->%X/%X "
			     "%s %hX %hX %hu "
			     "%u %u %u %u %hu-%hu %hu-%hu "
			     "A%hX X%hX %hX %u %hu %s"
			,tfw.label
			,&tfw.ipfw.fw_src.s_addr,&tfw.ipfw.fw_smsk.s_addr
			,&tfw.ipfw.fw_dst.s_addr,&tfw.ipfw.fw_dmsk.s_addr
			,tfw.ipfw.fw_vianame
			,&tfw.ipfw.fw_flg,	&tfw.ipfw.fw_invflg
			,&tfw.ipfw.fw_proto
			,&pkthi, &pktlo, &bytehi, &bytelo
			,&tfw.ipfw.fw_spts[0],	&tfw.ipfw.fw_spts[1]
			,&tfw.ipfw.fw_dpts[0],	&tfw.ipfw.fw_dpts[1]
			,&tfw.ipfw.fw_tosand,	&tfw.ipfw.fw_tosxor
			,&tfw.ipfw.fw_redirpt,	&tfw.ipfw.fw_mark
			,&tfw.ipfw.fw_outputsize,	&tfw.rule);

		if(set_and_quit) {
			sfw.ipfw = tfw.ipfw;
			if(!strcmp(tfw.rule, IP_FW_LABEL_NONE))
				strcpy(tfw.rule, "-");
			if(!append_entry(tfw.label, tfw.rule, sfw)) {
				fclose(fp);
				return;
			}
		} else
			put_to_rulelist(tfw);

	}
	if(filename)
		get_masq_timeout(fp);
	else
		get_masq_timeout(NULL);

	fclose(fp);
}

static void put_to_rulelist(struct _fwrule tfw)
{
	GtkCList *tlist;
	gint i, idx=0, row;
	gchar protocol[64];

	get_ipaddr(fbuf, tfw.ipfw.fw_src.s_addr, tfw.ipfw.fw_smsk.s_addr);
	content[idx++] = content_new(fbuf);

	proto_itos(protocol, tfw.ipfw.fw_proto);
	get_port(fbuf, tfw.ipfw.fw_spts[0], tfw.ipfw.fw_spts[1], protocol);
	content[idx++] = content_new(fbuf);

	get_ipaddr(fbuf, tfw.ipfw.fw_dst.s_addr, tfw.ipfw.fw_dmsk.s_addr);
	content[idx++] = content_new(fbuf);
		
	get_port(fbuf, tfw.ipfw.fw_dpts[0], tfw.ipfw.fw_dpts[1], protocol);
	content[idx++] = content_new(fbuf);

	content[idx++] = content_new(protocol);
		
	if(tfw.ipfw.fw_vianame[0] == '-')
		tfw.ipfw.fw_vianame[0] = '\0';
	content[idx++] = content_new(tfw.ipfw.fw_vianame);

	if(tfw.rule[0] == '-')
		strcpy(tfw.rule, IP_FW_LABEL_NONE);
	content[idx++] = content_new(tfw.rule);

	sprintf(fbuf,"%02X",tfw.ipfw.fw_tosand);
	content[idx++] = content_new(fbuf);

	sprintf(fbuf,"%02X",tfw.ipfw.fw_tosxor);
	content[idx++] = content_new(fbuf);

	put_unvisible_data(tfw);

	for(i=0; fwchain[i].label[0]; i++) {
		if(!strcmp(fwchain[i].label, tfw.label))
				break;
	}
	if(strcmp(tfw.rule, IP_FW_LABEL_ACCEPT) &&
	   strcmp(tfw.rule, IP_FW_LABEL_BLOCK) &&
	   strcmp(tfw.rule, IP_FW_LABEL_MASQUERADE) &&
	   strcmp(tfw.rule, IP_FW_LABEL_REJECT) &&
	   strcmp(tfw.rule, IP_FW_LABEL_REDIRECT) &&
	   strcmp(tfw.rule, IP_FW_LABEL_NONE) && fwchain[i].label[0]) {
		tlist = GTK_CLIST(fwchain[i].chainlist);
	} else if(fwchain[i].label[0]) {
		tlist = GTK_CLIST(fwchain[i].rulelist);
	}
	row = gtk_clist_append(tlist, content);
	if(tlist == GTK_CLIST(fwchain[i].rulelist))
		inverse_mark(tlist, row, tfw.ipfw.fw_invflg);
		
	for(i=0; i<NUMCOLS; i++)
		g_free(content[i]);
}

#include "Pics/inverse.xpm"
#include "Pics/normal.xpm"
void inverse_mark(GtkCList *tlist, gint row, __u16 fw_invflg)
{
	__u16 invflag[] = {
		IP_FW_INV_SRCIP,
		IP_FW_INV_SRCPT,
		IP_FW_INV_DSTIP,
		IP_FW_INV_DSTPT,
		IP_FW_INV_PROTO,
		IP_FW_INV_VIA,
		0
	};
	gint i;
	GdkPixmap *pixmap, *invpixmap, *nrmpixmap;
	GdkBitmap *mask, *invmask, *nrmmask;
	gchar **data, *text;
	
	invpixmap = gdk_pixmap_create_from_xpm_d(window->window, &invmask,
			&notebook->style->bg[GTK_STATE_NORMAL], inverse);
	nrmpixmap = gdk_pixmap_create_from_xpm_d(window->window, &nrmmask,
			&notebook->style->bg[GTK_STATE_NORMAL], normal);

	for(i=0; invflag[i]; i++) {
		pixmap = (fw_invflg & invflag[i]) ? invpixmap : nrmpixmap;
		mask = (fw_invflg & invflag[i]) ? invmask : nrmmask;
		gtk_clist_get_text(tlist, row, i, &text);
		gtk_clist_set_pixtext(tlist, row, i,
				g_strdup(text), 0, pixmap, mask);
	}
	
	gdk_pixmap_unref(invpixmap);
	gdk_pixmap_unref(invmask);
	gdk_pixmap_unref(nrmpixmap);
	gdk_pixmap_unref(nrmmask);
}

static void put_unvisible_data(struct _fwrule tfw)
{
	gint i=9;
	gchar protocol[64];

	proto_itos(protocol, tfw.ipfw.fw_proto);

	port_itos(fbuf, tfw.ipfw.fw_spts[0], protocol);
	content[i++] = content_new(fbuf);

	port_itos(fbuf, tfw.ipfw.fw_spts[1], protocol);
	content[i++] = content_new(fbuf);

	port_itos(fbuf, tfw.ipfw.fw_dpts[0], protocol);
	content[i++] = content_new(fbuf);

	port_itos(fbuf, tfw.ipfw.fw_dpts[1], protocol);
	content[i++] = content_new(fbuf);

	sprintf(fbuf, "%hu", tfw.ipfw.fw_flg);
	content[i++] = content_new(fbuf);

	sprintf(fbuf, "%hu", tfw.ipfw.fw_invflg);
	content[i++] = content_new(fbuf);

	sprintf(fbuf, "%hu", tfw.ipfw.fw_redirpt);
	content[i++] = content_new(fbuf);

	sprintf(fbuf, "%u", tfw.ipfw.fw_mark);
	content[i++] = content_new(fbuf);

	sprintf(fbuf, "%hu", tfw.ipfw.fw_outputsize);
	content[i++] = content_new(fbuf);

	content[i++] = content_new("0");
}

void show_interface_desc(gint idx, gint flag)
{
	gint i, j;
	gchar dbuf[128], *via;
	GtkWidget *button;
	GtkWidget *list;
	GList *sel;
	
	if(fwchain[idx].policy[0] != '-')
		return;

	if(find_chain(&i, &j, fwchain[idx].label)) {
		if(fwchain[idx].interface_desc)
			gtk_widget_destroy(fwchain[idx].interface_desc);
		gtk_clist_get_text(GTK_CLIST(fwchain[i].chainlist),j,5,&via);

		sprintf(dbuf, "CHAIN: %s - INTERFACE: %s",
					fwchain[i].label, via);
		fwchain[idx].interface_desc = gtk_label_new(dbuf);
		gtk_box_pack_start(GTK_BOX(fwchain[idx].hbox),
				fwchain[idx].interface_desc, FALSE, FALSE, 2);
		gtk_widget_show(fwchain[idx].interface_desc);
	}
	if(flag >= 0)
		return;

	button = gtk_button_new_with_label(" Edit Chain ");
	gtk_box_pack_end(GTK_BOX(fwchain[idx].hbox), button, FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(edit_chain_window),(ChainAction *)CHILD);
	gtk_widget_show(button);
	button = gtk_check_button_new_with_label(" Delete Chain ");
	gtk_box_pack_end(GTK_BOX(fwchain[idx].hbox), button, FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button), "toggled",
		GTK_SIGNAL_FUNC(delete_chain), NULL);
	gtk_widget_show(button);
}

static void clearChain()
{
	gint i;

	maxchains = 0;
	for(i=0; fwchain[i].label[0]; i++) {
		gtk_notebook_remove_page(GTK_NOTEBOOK(notebook), 0);
	}
	memset(fwchain, 0x00, sizeof(fwchain));
}

static void force_load_file(GtkWidget *widget, gpointer data)
{
	modified = 0;
	load_from_file(widget, data);
}

void load_from_file(GtkWidget *widget, gpointer data)
{
	if(modified) {
		dialog_window("Data has changed. continue ?",
				GTK_SIGNAL_FUNC(force_load_file));
		return;
	}

	openpurpose = FILE_OPEN;
	
	file_select();
}

static void force_load_system(GtkWidget *widget, gpointer data)
{
	modified = 0;

	load_from_system(widget, data);
}

void load_from_system(GtkWidget *widget, gpointer data)
{
	if(modified) {
		dialog_window("Data has changed. continue ?",
				GTK_SIGNAL_FUNC(force_load_system));
		return;
	}

	get_fw_data(NULL, 0);
}

void save_to_file(void)
{
	openpurpose = FILE_SAVE;

	if(current_filename) {
		save_fw_data(NULL, current_filename);
		modified = 0;
		return;
	}

	file_select();
}

void save_as_to_file(void)
{
	openpurpose = FILE_SAVE;

	modified = 0;
	file_select();
}

static void file_select(void)
{
	static GtkWidget *filew = NULL;

	if(!filew) {
		filew = gtk_file_selection_new("File selection");
		gtk_window_set_position(GTK_WINDOW(filew), GTK_WIN_POS_CENTER);
		gtk_signal_connect(GTK_OBJECT(filew), "destroy",
			GTK_SIGNAL_FUNC(file_destroy), filew);
		gtk_signal_connect(GTK_OBJECT(filew), "delete_event",
			GTK_SIGNAL_FUNC(file_destroy), filew);
		gtk_signal_connect(
			GTK_OBJECT(GTK_FILE_SELECTION(filew)->ok_button),
			"clicked", GTK_SIGNAL_FUNC(file_ok_sel), filew);
		gtk_signal_connect(
			GTK_OBJECT(GTK_FILE_SELECTION(filew)->cancel_button),
			"clicked", GTK_SIGNAL_FUNC(file_can_sel), filew);
	}
	gtk_widget_show(filew);
}

static void file_ok_sel(GtkWidget *w, GtkWidget *fs)
{
	gchar *fname;
	GtkFileSelection *filesel;
	
	filesel = GTK_FILE_SELECTION(fs);
	
	fname = gtk_file_selection_get_filename(filesel);
	gtk_widget_hide(fs);
	if(!fname)
		return;
	if(fname[strlen((const char *)fname)-1] == '/')
		return;

	switch(openpurpose) {
	case FILE_OPEN : get_fw_data(fname, 0); break;
	case FILE_SAVE : save_fw_data(NULL, fname); break;
	default : break;
	}
}

static void file_can_sel(GtkWidget *w, GtkWidget *fs)
{
	gtk_widget_hide(fs);
}

static void file_destroy(GtkWidget *w, GtkWidget *fs)
{
	gtk_widget_hide(w);
}

GtkWidget *menu_new;
GtkWidget *menu_edit;
GtkWidget *menu_insert;
GtkWidget *menu_copy;
GtkWidget *menu_cut;
GtkWidget *menu_paste;
GtkWidget *menu_delete;
GtkWidget *menu_duplicate;

static GtkWidget *create_rule_popup_menu(void)
{
	GtkWidget *popmenu;
	GtkWidget *menu = gtk_menu_new();
	
	menu_new = create_menu_item(menu, " New  ", TRUE,
		GTK_SIGNAL_FUNC(popup_rule_edit),(RuleAction *)NEW);
	menu_edit = create_menu_item(menu, " Edit  ", FALSE,
		GTK_SIGNAL_FUNC(popup_rule_edit),(RuleAction *)EDIT);
	menu_insert = create_menu_item(menu, " Insert  ", FALSE,
		GTK_SIGNAL_FUNC(popup_rule_edit),(RuleAction *)INSERT);
	menu_copy = create_menu_item(menu, " Copy  ", FALSE,
			GTK_SIGNAL_FUNC(popup_rule_copy), NULL);
	menu_cut = create_menu_item(menu, " Cut  ", FALSE,
			GTK_SIGNAL_FUNC(popup_rule_cut), NULL);
	menu_paste = create_menu_item(menu, " Paste  ", FALSE,
			GTK_SIGNAL_FUNC(popup_rule_paste), NULL);
	menu_delete = create_menu_item(menu, " Delete  ", FALSE,
			GTK_SIGNAL_FUNC(popup_rule_delete), NULL);
	menu_duplicate = create_menu_item(menu, " Duplicate  ", FALSE,
			GTK_SIGNAL_FUNC(popup_rule_duplicate), NULL);
	menu_separator(menu);
	popmenu = create_menu_item(menu, " Quit ", TRUE,
			GTK_SIGNAL_FUNC(program_quit), NULL);
	
	return menu;
}

static gint rulelist_event_handler(
		GtkWidget *w, GdkEvent *event, gpointer data)
{
	GdkEventButton *bevent;
	GList *selection;
	gint page, active, multi, paste;

	switch(event->type) {
	case GDK_BUTTON_PRESS:
		switch(event->button.button) {
		case 3:
			page = gtk_notebook_get_current_page(
					GTK_NOTEBOOK(notebook));
			selection =
				GTK_CLIST(fwchain[page].rulelist)->selection;
			active = selection ? TRUE : FALSE;
			multi = (selection && selection->next) ? TRUE : FALSE;
			paste = GTK_CLIST(cliplist)->rows ? TRUE : FALSE;

			gtk_widget_set_sensitive(menu_edit, (active&&!multi));
			gtk_widget_set_sensitive(menu_insert, (active&&!multi));
			gtk_widget_set_sensitive(menu_copy, active);
			gtk_widget_set_sensitive(menu_cut, active);
			gtk_widget_set_sensitive(menu_paste, paste);
			gtk_widget_set_sensitive(menu_delete, active);
			gtk_widget_set_sensitive(
					menu_duplicate, (active&&!multi));

			bevent = (GdkEventButton *)event;
			gtk_menu_popup(GTK_MENU(popup_menu), NULL, NULL,
				NULL, NULL, bevent->button, bevent->time);
			return TRUE;
		default: break;
		}
		break;
	case GDK_2BUTTON_PRESS:
		switch(event->button.button) {
		case 1:
			popup_rule_edit(w, EDIT);
			return TRUE;
		default: break;
		}
		break;
	default: break;
	}
	return FALSE;
}

static void set_fwpolicy_data(GtkWidget *widget, gpointer data)
{
	gint page = gtk_notebook_current_page(GTK_NOTEBOOK(notebook));

	if(page < 0)
		return;
	if(!strcmp(fwchain[page].policy, data))
		return;
	
	modified = 1;
	strcpy(fwchain[page].policy, data);
}

void save_fw_data(GtkWidget *widget, gchar *filename)
{
	FILE *fp;
	guint i, row=0;
	struct ip_fwuser sfw;
	ip_chainlabel rule;
	gchar *text, buf[64];
	
	if(filename) {
		fp = fopen(filename, "w+");
		if(!fp) {
			sprintf(fbuf, "File %s open error", filename);
			dialog_window(fbuf, NULL);
			return;
		}
		fputs("#!ip-firewall rule  *DO NOT EDIT*\n", fp);
	} else {
		flush_entries(NULL);
		delete_user_chain(NULL);
	}

	for(i=0; i<maxchains; i++) {
		if(fwchain[i].deleted)
			continue;
		if(filename) {
			fprintf(fp, "%s %s %u %u %u %u %u\n",
				fwchain[i].label, fwchain[i].policy,
				fwchain[i].refcnt, 0,0,0,0);
		} else if(strcmp(fwchain[i].label, IP_FW_LABEL_INPUT) &&
			  strcmp(fwchain[i].label, IP_FW_LABEL_FORWARD) &&
			  strcmp(fwchain[i].label, IP_FW_LABEL_OUTPUT)) {
			ipfwc_create_chain(fwchain[i].label);
		} else {
			ipfwc_set_policy(fwchain[i].label, fwchain[i].policy);
		}
	}
	if(filename)
		fputs("*\n", fp);
	
	for(i=0; i<maxchains; i++) {
		if(fwchain[i].deleted)
			continue;
		strcpy(sfw.label, fwchain[i].label);
		for(row=0; row<GTK_CLIST(fwchain[i].rulelist)->rows; row++) {
			conv_fw_data(&sfw, rule,
				GTK_CLIST(fwchain[i].rulelist), row, TRUE);
			if(filename)
				write_to_file(fp, sfw, rule);
			else if(!dialog_append_entry(
					fwchain[i].label, rule, sfw)) {
				gtk_notebook_set_page(GTK_NOTEBOOK(notebook),i);
				gtk_clist_unselect_all(
					GTK_CLIST(fwchain[i].rulelist));
				gtk_clist_select_row(
					GTK_CLIST(fwchain[i].rulelist),row,0);
				return;
			}
		}
		for(row=0; row<GTK_CLIST(fwchain[i].chainlist)->rows; row++) {
			gtk_clist_get_text(GTK_CLIST(fwchain[i].chainlist),
						row, 18, &text);
			if(*text == '1')
				continue;
			conv_fw_data(&sfw, rule,
				GTK_CLIST(fwchain[i].chainlist), row, FALSE);
			if(filename)
				write_to_file(fp, sfw, rule);
			else if(!dialog_append_entry(
						fwchain[i].label, rule, sfw))
				return;
		}
	}
	
	if(filename && save_masq_timeout(fp)) {
		if(strcmp(filename, GFCC_HOME"/system.rule")) {
			sprintf(buf, "Successfully Saved to\n%s", filename);
			dialog_window(buf, NULL);
		}
		fclose(fp);
	} else if(set_masq_timeout(NULL, NULL)) {
		dialog_window("Successfully Applied to system", NULL);
		save_fw_data(NULL, GFCC_HOME"/system.rule");
	}
	
	return;
}

static void write_to_file(FILE *fp, struct ip_fwuser sfw, gchar *rule)
{
	fprintf(fp, "%9s "
		    "%08lX/%08lX->%08lX/%08lX "
		    "%s "
		    "%hX %hX "
		    "%hu "
		    "%-9lu %-9lu %-9lu %-9lu "
		    "%hu-%hu %hu-%hu "
		    "A%02X X%02X "
		    "%08X "
		    "%hu %hu"
		    "%10s\n",
		    sfw.label,
		    sfw.ipfw.fw_src.s_addr,
		    sfw.ipfw.fw_smsk.s_addr,
		    sfw.ipfw.fw_dst.s_addr,
		    sfw.ipfw.fw_dmsk.s_addr,
		    sfw.ipfw.fw_vianame,
		    sfw.ipfw.fw_flg, sfw.ipfw.fw_invflg,
		    sfw.ipfw.fw_proto,
		    0, 0, 0, 0,
		    sfw.ipfw.fw_spts[0], sfw.ipfw.fw_spts[1],
		    sfw.ipfw.fw_dpts[0], sfw.ipfw.fw_dpts[1],
		    sfw.ipfw.fw_tosand, sfw.ipfw.fw_tosxor,
		    sfw.ipfw.fw_redirpt, sfw.ipfw.fw_mark,
		    sfw.ipfw.fw_outputsize, rule);
}

void get_service_name()
{
	FILE *fp;
	gint tcp=0, udp=0;
	guint portval;
	gchar portlabel[32], proto[8];
	
	fp = fopen("/etc/services", "r");
	if(!fp)
		return;
	
	while(fgets(fbuf, sizeof(fbuf), fp)) {
		if(fbuf[0] == '#' || fbuf[0] == '\n')
			continue;
		sscanf(fbuf, "%s %u/%s", portlabel, &portval, proto);
		if(!strcmp(proto, "tcp")) {
			service_tcp[tcp].label = g_malloc(strlen(portlabel)+1);
			strcpy(service_tcp[tcp].label, portlabel);
			service_tcp[tcp++].value = portval;
		} else if(!strcmp(proto, "udp")) {
			service_udp[udp].label = g_malloc(strlen(portlabel)+1);
			strcpy(service_udp[udp].label, portlabel);
			service_udp[udp++].value = portval;
		}
	}
	service_tcp[tcp].label = NULL;
	service_udp[udp].label = NULL;

	fclose(fp);
}

void get_proto_name()
{
	FILE *fp;
	gint i=0;
	guint protoval;
	gchar protolabel[32];
	
	fp = fopen("/etc/protocols", "r");
	if(!fp) {
		perror("/etc/protocols");
		return;
	}
	
	while(fgets(fbuf, sizeof(fbuf), fp)) {
		if(fbuf[0] == '#' || fbuf[0] == '\n')
			continue;
		sscanf(fbuf, "%s %u", protolabel, &protoval);
		if(!protoval)
			strcpy(protolabel, "all");

		proto_name[i].label = g_malloc(strlen(protolabel)+1);
		strcpy(proto_name[i].label, protolabel);
		proto_name[i++].value = protoval;
	}
	proto_name[i].label = NULL;

	fclose(fp);
}
