#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>

#include "gfcc.h"

extern struct _fwchain fwchain[MAXCHAINS];
extern struct _menu service_tcp[];
extern struct _menu service_udp[];
extern struct _menu proto_name[];
extern struct _menu tos_name[];
extern gint maxchains;
extern gint modified;

static guint get_mask(guint);
static void close_dialog(GtkWidget *, gpointer);

void (* tcallback)();

long int_from_hex(gchar *data)
{
	long Result;
	gint Hi, Lo;
	
	if(!*data)
		return (__u8)0;
	
	Hi = *data;
	if(Hi >= '0' && Hi <= '9') {
		Hi -= '0';
	} else if(Hi >= 'a' && Hi <= 'f') {
		Hi -= ('a'-10);
	} else if(Hi >= 'A' && Hi <= 'F') {
		Hi -= ('A'-10);
	}
	
	Lo = *(data+1);
	if(Lo >= '0' && Lo <= '9') {
		Lo -= '0';
	} else if(Lo >= 'a' && Lo <= 'f') {
		Lo -= ('a'-10);
	} else if(Lo >= 'A' && Lo <= 'F') {
		Lo -= ('A'-10);
	}

	Result = Lo + (16 * Hi);

	return Result;
}

void tosand_itos(gchar *buf, __u8 tosand)
{
	gint i;
	
	for(i=0; tos_name[i].label; i++) {
		if(tosand == (__u8)tos_name[i].value) {
			strcpy(buf, tos_name[i].label);
			return;
		}
	}
}

__u8 tosand_stoi(gchar *data)
{
	gint i;
	
	for(i=0; tos_name[i].label; i++) {
		if(!strcmp(tos_name[i].label, data)) {
			return (__u8)tos_name[i].value;
		}
	}
	return (__u8)0xff;
}

void proto_itos(gchar *data, guint proto)
{
	gint i;
	
	for(i=0; proto_name[i].label; i++) {
		if(proto == proto_name[i].value) {
			strcpy(data, proto_name[i].label);
			return;
		}
	}
	sprintf(data, "%u", proto);
	return;
}

guint proto_stoi(gchar *data)
{
	gint i;
	
	for(i=0; proto_name[i].label; i++) {
		if(!strcmp(data, proto_name[i].label)) {
			return proto_name[i].value;
		}
	}
	return (guint)atoi(data);
}

void port_itos(gchar *data, guint port, gchar *proto)
{
	gint i;
	struct _menu *tmp;
	
	if(!strcmp(proto, "tcp"))
		tmp = service_tcp;
	else if(!strcmp(proto, "udp"))
		tmp = service_udp;
	else {
		sprintf(data, "%u", port);
		return;
	}

	
	for(; tmp->label; tmp++) {
		if(port == (guint)tmp->value) {
			strcpy(data, tmp->label);
			return;
		}
	}
	
	sprintf(data, "%u", port);
	return;
}

guint port_stoi(gchar *data)
{
	gint i;
	
	for(i=0; service_tcp[i].label; i++) {
		if(!strcmp(data, service_tcp[i].label)) {
			return (guint)service_tcp[i].value;
		}
	}
	for(i=0; service_udp[i].label; i++) {
		if(!strcmp(data, service_udp[i].label)) {
			return (guint)service_udp[i].value;
		}
	}
	
	return (guint)atoi(data);
}

void get_port(gchar *buf, guint pmin, guint pmax, gchar *proto)
{
	gchar bmin[10], bmax[10];

	port_itos(bmin, pmin, proto);
	port_itos(bmax, pmax, proto);
	if(pmin != pmax) {
		if(pmin || pmax < 65535) {
			sprintf(buf, "%s:%s", bmin, bmax);
		} else
			strcpy(buf, "n/a");
	} else
		sprintf(buf, "%s", bmin);
}

static void close_dialog(GtkWidget *widget, gpointer data)
{
	gtk_widget_destroy(gtk_widget_get_toplevel(widget));
	tcallback();
}

void dialog_window(gchar *message, GtkSignalFunc callback)
{
	GtkWidget *dwin, *label, *button;
	
	dwin = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dwin), "Message Window");
	gtk_window_set_position(GTK_WINDOW(dwin), GTK_WIN_POS_CENTER);
	gtk_container_set_border_width(GTK_CONTAINER(dwin), 0);
	gtk_widget_set_usize(dwin, 250, 150);
	
	label = gtk_label_new(message);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dwin)->vbox), label,TRUE,TRUE,0);
	gtk_widget_show(label);
	
	if(callback) {
		tcallback = callback;
		button = gtk_button_new_with_label("OK");
		gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(close_dialog), GTK_OBJECT(dwin));
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dwin)->action_area),
				button, TRUE, TRUE, 0);
		gtk_widget_show(button);

		button = gtk_button_new_with_label("Cancel");
	} else
		button = gtk_button_new_with_label("OK");
	
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dwin));
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dwin)->action_area), button,
			TRUE, TRUE, 0);
	gtk_widget_show(button);
	
	gtk_widget_show(dwin);
}

gint find_chain(gint *page, gint *row, gchar *clabel)
{
	GtkCList *tlist;
	gchar *parent;

	for((*page)=0; *page<maxchains; (*page)++) {
		if(strcmp(fwchain[*page].label, IP_FW_LABEL_INPUT) &&
		   strcmp(fwchain[*page].label, IP_FW_LABEL_OUTPUT) &&
		   strcmp(fwchain[*page].label, IP_FW_LABEL_FORWARD))
			continue;

		tlist = GTK_CLIST(fwchain[*page].chainlist);
		for((*row)=0; *row < tlist->rows; (*row)++) {
			gtk_clist_get_text(tlist, *row, 6, &parent);

			if(!strcmp(parent, clabel))
				return 1;
		}
	}
	return 0;
}

GtkWidget *new_pixmap_button(GtkWidget *box, gchar **data)
{
	GtkWidget *button;
	GtkWidget *pixmap;

	button = gtk_button_new();
	gtk_box_pack_start(GTK_BOX(box), button, FALSE, FALSE, 0);
	gtk_widget_show(button);
	gtk_widget_realize(button);

	pixmap = new_pixmap(button, data);
	gtk_container_add(GTK_CONTAINER(button), pixmap);
	gtk_widget_show(pixmap);

	return button;
}

GtkWidget *new_pixmap(GtkWidget *w, gchar **data)
{
	GdkPixmap *gdkpixmap;
	GdkBitmap *pmask;
	GtkWidget *pixmap;
	
	gdkpixmap = gdk_pixmap_create_from_xpm_d(w->window, &pmask,
			&w->style->bg[GTK_STATE_NORMAL], data);
	pixmap = gtk_pixmap_new(gdkpixmap, pmask);
	
	gdk_pixmap_unref(gdkpixmap);
	gdk_pixmap_unref(pmask);
	
	return pixmap;
}

void force_quit(GtkWidget *widget, gpointer data)
{
	modified = 0;
	program_quit(widget, NULL, NULL);
}

void program_quit(GtkWidget *widget, GtkWidget *event, gpointer data)
{
	if(modified) {
		dialog_window("Data has changed. continue ?",
				GTK_SIGNAL_FUNC(force_quit));
		return;
	}
	save_host_name(NULL, NULL);
	gtk_main_quit();
}

GtkWidget * append_submenu_item(
		GtkWidget *submenu, GtkSignalFunc func, gchar *text)
{
	GtkWidget *menuitem = gtk_menu_item_new_with_label(text);
	gtk_widget_show(menuitem);
	gtk_menu_append(GTK_MENU(submenu), menuitem);
	
	gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
			GTK_SIGNAL_FUNC(func), text);

	return menuitem;
}

void get_ipaddr(gchar *buf, __u32 ipaddr, __u32 mask)
{
	gchar ipbuf[64];

	if(!ipaddr) {
		strcpy(buf, "Any");
		return;
	}
	sprintf(ipbuf, "%u.%u.%u.%u/%u",
		               (ipaddr >> 24) & 0xff,
		               (ipaddr >> 16) & 0xff,
		               (ipaddr >>  8) & 0xff,
			       (ipaddr & 0xff), get_mask(mask));
	host_iptoname(buf, ipbuf);
}

static guint get_mask(guint maskval)
{
	gint i, mask=0;

	for(i=32; i>0; i--) {
		if(!((maskval >> i-1) & 0x01))
			break;
		mask++;
	}
	
	return mask;
}

GtkWidget *create_menu_item(GtkWidget *menu, gchar *label,
	gint sensitive_flag, GtkSignalFunc this_func, gpointer this_func_data)
{
	GtkWidget *menu_item = gtk_menu_item_new_with_label(label);
	gtk_menu_append(GTK_MENU(menu), menu_item);
	gtk_widget_set_sensitive(menu_item, sensitive_flag);
	gtk_widget_show(menu_item);
	gtk_signal_connect(GTK_OBJECT(menu_item), "activate",
			GTK_SIGNAL_FUNC(this_func), this_func_data);
	
	return menu_item;
}

void menu_separator(GtkWidget *menu)
{
	GtkWidget *menu_item = gtk_menu_item_new();
	gtk_menu_append(GTK_MENU(menu), menu_item);
	gtk_widget_set_sensitive(menu_item, FALSE);
	gtk_widget_show(menu_item);
}

gchar *content_new(gchar *buf)
{
	gchar *tmp;
	
	tmp = g_malloc(strlen(buf)+1);
	if(!tmp)
		return NULL;

	strcpy(tmp, buf);
	
	return tmp;
}
